/****************************************************************/
// The Maze Project v1.0
// Jason Yanchuleff
// Last Revised: 05.01.06
//
// Main application code for the maze project.
/****************************************************************/

#include <windows.h>												// Header File For Windows
#include <stdio.h>													// Header File For Standard Input/Output
#include <string.h>
#include <commctrl.h>
#include <gl\gl.h>													// Header File For The OpenGL32 Library
#include <gl\glu.h>													// Header File For The GLu32 Library
#include <gl\glaux.h>												// Header File For The Glaux Library
#include "stdafx.h"
#include <sapi.h>

#include "fmod.hpp"
#include "fmod_errors.h"

#include "MazeApp.h"												// Header File For Maze Application
#include "Joystick.h"												// Header File For Force Feedback Joystick

HDC			hDC=NULL;												// Private GDI Device Context
HGLRC		hRC=NULL;												// Permanent Rendering Context
HWND		hWnd=NULL;												// Holds Our Window Handle
//HWND		hWndStatus=NULL;										// Holds Our Status Bar Handle
HINSTANCE	hInstance;												// Holds The Instance Of The Application

MazeApp *MAZEGAME = new MazeApp();									// Main Application Object
Joystick *tJoystick;

bool	keys[256];													// Array Used For The Keyboard Routine
bool	active=TRUE;												// Window Active Flag Set To TRUE By Default
bool	fullscreen=TRUE;											// Fullscreen Flag Set To Fullscreen Mode By Default
bool	use_joystick=FALSE;

double js_pos[3];
bool rotated = false;
int rstate = 2;
int rinc = 0;
FMOD_VECTOR pos;
FMOD_VECTOR forward;

char	szBuf[100];

HRESULT hr=NULL;
ISpVoice * pVoice = NULL;

bool ended = false;

GLfloat LightAmbient[]= { 0.5f, 0.5f, 0.5f, 1.0f }; 				// Ambient Light Values
GLfloat LightDiffuse[]= { 1.0f, 1.0f, 1.0f, 1.0f };					// Diffuse Light Values
GLfloat LightPosition[]= { 0.0f, 0.0f, 2.0f, 1.0f };				// Light Position

GLfloat fogColor[4]= {0.5f, 0.5f, 1.0f, 1.0f};						// Fog Color

GLuint	base;														// Base Display List For The Font Set
GLYPHMETRICSFLOAT gmf[256];

AUX_RGBImageRec *rImage;

const float pi_180 = 0.0174532925f;
clock_t time_now, movement_timer = 0;

LRESULT	CALLBACK WndProc(HWND, UINT, WPARAM, LPARAM);				// Declaration For WndProc
void Normalize(float &a, float &b, float &c);

AUX_RGBImageRec *LoadBMP(const char *Filename)						// Loads A Bitmap Image
{
	FILE *File=NULL;												// File Handle

	if (!Filename)													// Make Sure A Filename Was Given
	{
		return NULL;												// If Not Return NULL
	}

	fopen_s(&File,Filename,"r");										// Check To See If The File Exists

	if (File)														// Does The File Exist?
	{
		fclose(File);												// Close The Handle
		return auxDIBImageLoad(Filename);							// Load The Bitmap And Return A Pointer
	}

	return NULL;													// If Load Failed Return NULL
}

GLuint LoadGLTexture( const char *filename )						// Load Bitmaps And Convert To Textures
{
	AUX_RGBImageRec *pImage;										// Create Storage Space For The Texture
	GLuint texture = 0;												// Texture ID

	pImage = LoadBMP( filename );									// Loads The Bitmap Specified By filename

	// Load The Bitmap, Check For Errors, If Bitmap's Not Found Quit
	if ( pImage != NULL && pImage->data != NULL )					// If Texture Image Exists
	{
		glGenTextures(1, &texture);									// Create The Texture

		// Typical Texture Generation Using Data From The Bitmap
		glBindTexture(GL_TEXTURE_2D, texture);
		glTexImage2D(GL_TEXTURE_2D, 0, 3, pImage->sizeX, pImage->sizeY, 0, GL_RGB, GL_UNSIGNED_BYTE, pImage->data);
		glTexParameteri(GL_TEXTURE_2D,GL_TEXTURE_MIN_FILTER,GL_LINEAR);
		glTexParameteri(GL_TEXTURE_2D,GL_TEXTURE_MAG_FILTER,GL_LINEAR);

		free(pImage->data);											// Free The Texture Image Memory
		free(pImage);												// Free The Image Structure
	}

	return texture;													// Return The Status
}

GLvoid ReSizeGLScene(GLsizei width, GLsizei height)					// Resize And Initialize The GL Window
{
	if (height==0)													// Prevent A Divide By Zero By
	{
		height=1;													// Making Height Equal One
	}

	glViewport(0,0,width,height);									// Reset The Current Viewport

	glMatrixMode(GL_PROJECTION);									// Select The Projection Matrix
	glLoadIdentity();												// Reset The Projection Matrix

	// Calculate The Aspect Ratio Of The Window
	gluPerspective(45.0f,(GLfloat)width/(GLfloat)height,1.0f,1000.0f);	// View Depth of 1000

	glMatrixMode(GL_MODELVIEW);										// Select The Modelview Matrix
	glLoadIdentity();												// Reset The Modelview Matrix
}

GLvoid BuildFont(GLvoid)					// Build Our Bitmap Font
{
	HFONT	font;						// Windows Font ID

	base = glGenLists(256);					// Storage For 256 Characters

	font = CreateFont(	-12,				// Height Of Font
				0,				// Width Of Font
				0,				// Angle Of Escapement
				0,				// Orientation Angle
				FW_BOLD,			// Font Weight
				FALSE,				// Italic
				FALSE,				// Underline
				FALSE,				// Strikeout
				ANSI_CHARSET,			// Character Set Identifier
				OUT_TT_PRECIS,			// Output Precision
				CLIP_DEFAULT_PRECIS,		// Clipping Precision
				ANTIALIASED_QUALITY,		// Output Quality
				FF_DONTCARE|DEFAULT_PITCH,	// Family And Pitch
				"Comic Sans MS");		// Font Name

	SelectObject(hDC, font);				// Selects The Font We Created

	wglUseFontOutlines(	hDC,				// Select The Current DC
				0,				// Starting Character
				255,				// Number Of Display Lists To Build
				base,				// Starting Display Lists
				0.0f,				// Deviation From The True Outlines
				0.2f,				// Font Thickness In The Z Direction
				WGL_FONT_POLYGONS,		// Use Polygons, Not Lines
				gmf);				// Address Of Buffer To Recieve Data
}

GLvoid KillFont(GLvoid)						// Delete The Font
{
 	glDeleteLists(base, 256);				// Delete All 256 Characters
}

GLvoid glPrint(const char *fmt, ...)				// Custom GL "Print" Routine
{
	float		length=0;				// Used To Find The Length Of The Text
	char		text[256];				// Holds Our String
	va_list		ap;					// Pointer To List Of Arguments

	if (fmt == NULL)					// If There's No Text
		return;						// Do Nothing

	va_start(ap, fmt);					// Parses The String For Variables
	    vsprintf(text, fmt, ap);				// And Converts Symbols To Actual Numbers
	va_end(ap);						// Results Are Stored In Text

	for (unsigned int loop=0;loop<(strlen(text));loop++)	// Loop To Find Text Length
	{
		length+=gmf[text[loop]].gmfCellIncX;		// Increase Length By Each Characters Width
	}

	glTranslatef(-length/2,0.0f,0.0f);			// Center Our Text On The Screen

	glPushAttrib(GL_LIST_BIT);				// Pushes The Display List Bits
	glListBase(base);					// Sets The Base Character to 0

	glCallLists(strlen(text), GL_UNSIGNED_BYTE, text);	// Draws The Display List Text
	glPopAttrib();						// Pops The Display List Bits
}

int InitGL(GLvoid)													// All Setup For OpenGL Goes Here
{
	MAZEGAME->tMaze->LoadTileTextures();							// Loads Model Textures

	glEnable(GL_TEXTURE_2D);										// Enable Texture Mapping ( NEW )
	glShadeModel(GL_SMOOTH);										// Enable Smooth Shading
	glClearColor(0.0f, 0.0f, 0.0f, 0.5f);							// Black Background
	glClearDepth(1.0f);												// Depth Buffer Setup
	glEnable(GL_DEPTH_TEST);										// Enables Depth Testing
	glDepthFunc(GL_LEQUAL);											// The Type Of Depth Testing To Do
	glHint(GL_PERSPECTIVE_CORRECTION_HINT, GL_NICEST);				// Really Nice Perspective Calculations

	glLightfv(GL_LIGHT1, GL_AMBIENT, LightAmbient);				// Setup The Ambient Light
	glLightfv(GL_LIGHT1, GL_DIFFUSE, LightDiffuse);				// Setup The Diffuse Light
	glLightfv(GL_LIGHT1, GL_POSITION,LightPosition);			// Position The Light
	glEnable(GL_LIGHT1);										// Enable Light One
	glEnable(GL_LIGHTING);

	glClearColor(0.5f,0.5f,1.0f,1.0f);			// We'll Clear To The Color Of The Fog
	glFogi(GL_FOG_MODE, GL_LINEAR);				// Fog Mode
	glFogfv(GL_FOG_COLOR, fogColor);			// Set Fog Color
	glFogf(GL_FOG_DENSITY, 0.65f);				// How Dense Will The Fog Be
	glHint(GL_FOG_HINT, GL_DONT_CARE);			// Fog Hint Value
	glFogf(GL_FOG_START, 55.0f);				// Fog Start Depth
	glFogf(GL_FOG_END, 85.0f);					// Fog End Depth
	glEnable(GL_FOG);

	BuildFont();

	MAZEGAME->triggerMenu();					// HACK to setup menu properly on startup
	MAZEGAME->triggerMenu();

	//rImage = LoadBMP("Data/Images/test.bmp");

	return TRUE;													// Initialization Went OK
}

int DrawGLScene(GLvoid)												// Here's Where We Do All The Drawing
{
	time_now = clock();
	if(time_now - movement_timer > CLK_TCK/75)						// Restricts to 75 fps
	{
		movement_timer = time_now;

		glClear(GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT);				// Clear The Screen And The Depth Buffer
		glLoadIdentity();
		
		MAZEGAME->Draw();											// Draw frame
		MAZEGAME->tMaze->fsystem->update();							// Update FMOD
		MAZEGAME->UpdateObjects();									// Update sound objects

		if(MAZEGAME->completed && !ended)							// Check for end of game (should be revised)
		{
			if(SUCCEEDED(hr)) {
				hr = pVoice->Speak(L"Congratulations! You have completed the maze.", 0, NULL);
			}
			ended = true;
		}

		if(use_joystick)						// Handles joystick events
		{
			tJoystick->Update();
			tJoystick->GetPosition(js_pos);

			float move_x = 0.0f;
			float move_z = 0.0f;

			if(js_pos[0] >= 0.2f || js_pos[0] <= -0.2f)
			{
				move_x += (float)sin((MAZEGAME->tPlayer->heading+90)*pi_180) * (float)js_pos[0] * 0.75f;
				move_z += (float)cos((MAZEGAME->tPlayer->heading+90)*pi_180) * (float)js_pos[0] * 0.75f;

				if (MAZEGAME->tPlayer->walkbiasangle <= 1.0f)
				{
					MAZEGAME->tPlayer->walkbiasangle = 359.0f;
				}
				else
				{
					MAZEGAME->tPlayer->walkbiasangle-= 5;
				}
				MAZEGAME->tPlayer->walkbias = (float)sin(MAZEGAME->tPlayer->walkbiasangle * pi_180)/2.0f;
			}
			if(js_pos[1] >= 0.3f || js_pos[1] <= -0.3f)
			{
				move_x += (float)sin(MAZEGAME->tPlayer->heading*pi_180) * (float)js_pos[1] * 0.75f;
				move_z += (float)cos(MAZEGAME->tPlayer->heading*pi_180) * (float)js_pos[1] * 0.75f;

				if (MAZEGAME->tPlayer->walkbiasangle <= 1.0f)
				{
					MAZEGAME->tPlayer->walkbiasangle = 359.0f;
				}
				else
				{
					MAZEGAME->tPlayer->walkbiasangle-= 5;
				}
				MAZEGAME->tPlayer->walkbias = (float)sin(MAZEGAME->tPlayer->walkbiasangle * pi_180)/2.0f;
			}

			if(js_pos[2] > 0.2f && !rotated)				// Mess of logic to do 90 deg turn animation when the joystick is rotated
			{
				rotated = true;
				rstate = -1;
			}
			else if(js_pos[2] < -0.2f && !rotated)
			{
				rotated = true;
				rstate = 1;
			}
			else
			{
				if(rstate==2 && js_pos[2]==0.0f)
				{
					rotated = false;
				}
			}

			if(rstate==-1)
			{
				MAZEGAME->tPlayer->heading -= 10.0f;
				rinc++;
			}
			else if(rstate==1)
			{
				MAZEGAME->tPlayer->heading += 10.0f;
				rinc++;
			}
			if(rinc==9)
			{
				rstate = 2;
				rinc = 0;
			}

			bool bump = false;							// Handles force-feedback

			if(!MAZEGAME->DetectCollision(move_x, 0.0f))
				MAZEGAME->tPlayer->x += move_x;
			else
				bump = true;

			if(!MAZEGAME->DetectCollision(0.0f, move_z))
				MAZEGAME->tPlayer->z += move_z;
			else
				bump = true;

			if(bump)
				tJoystick->PlayBump(js_pos[0], js_pos[1]);

			pos.x = MAZEGAME->tPlayer->x;			// Move player and update FMOD
			pos.y = 5.0f;
			pos.z = MAZEGAME->tPlayer->z;
			forward.x = (float)sin(MAZEGAME->tPlayer->heading*pi_180);
			forward.y = 0.0f;
			forward.z = (float)cos(MAZEGAME->tPlayer->heading*pi_180);
			Normalize(forward.x, forward.y, forward.z);
			MAZEGAME->tMaze->fsystem->set3DListenerAttributes(0, &pos, 0, &forward, 0);
		}
	}

	/*int x = MAZEGAME->tPlayer->x;
	int z = MAZEGAME->tPlayer->z;
	wsprintf(szBuf, "Local Player Coords: %d, %d    Number of Collisions: %d", x, z, MAZEGAME->num_collisions);
	//SendMessage(hWndStatus, SB_SETTEXT, 0, (LPARAM)(LPSTR)szBuf);*/

	return TRUE;
}

GLvoid KillGLWindow(GLvoid)											// Properly Kill The Window
{
	if (fullscreen)													// Are We In Fullscreen Mode?
	{
		ChangeDisplaySettings(NULL,0);								// If So Switch Back To The Desktop
		ShowCursor(TRUE);											// Show Mouse Pointer
	}

	if (hRC)														// Do We Have A Rendering Context?
	{
		if (!wglMakeCurrent(NULL,NULL))								// Are We Able To Release The DC And RC Contexts?
		{
			MessageBox(NULL,"Release Of DC And RC Failed.","SHUTDOWN ERROR",MB_OK | MB_ICONINFORMATION);
		}

		if (!wglDeleteContext(hRC))									// Are We Able To Delete The RC?
		{
			MessageBox(NULL,"Release Rendering Context Failed.","SHUTDOWN ERROR",MB_OK | MB_ICONINFORMATION);
		}
		hRC=NULL;													// Set RC To NULL
	}

	if (hDC && !ReleaseDC(hWnd,hDC))								// Are We Able To Release The DC
	{
		MessageBox(NULL,"Release Device Context Failed.","SHUTDOWN ERROR",MB_OK | MB_ICONINFORMATION);
		hDC=NULL;													// Set DC To NULL
	}

	if (hWnd && !DestroyWindow(hWnd))								// Are We Able To Destroy The Window?
	{
		MessageBox(NULL,"Could Not Release hWnd.","SHUTDOWN ERROR",MB_OK | MB_ICONINFORMATION);
		hWnd=NULL;													// Set hWnd To NULL
	}

	if (!UnregisterClass("OpenGL",hInstance))						// Are We Able To Unregister Class
	{
		MessageBox(NULL,"Could Not Unregister Class.","SHUTDOWN ERROR",MB_OK | MB_ICONINFORMATION);
		hInstance=NULL;												// Set hInstance To NULL
	}

	KillFont();
}

BOOL CreateGLWindow(char* title, int width, int height, int bits, bool fullscreenflag)
{
	GLuint		PixelFormat;										// Holds The Results After Searching For A Match
	WNDCLASS	wc;													// Windows Class Structure
	DWORD		dwExStyle;											// Window Extended Style
	DWORD		dwStyle;											// Window Style
	RECT		WindowRect;											// Grabs Rectangle Upper Left / Lower Right Values
	WindowRect.left=(long)0;										// Set Left Value To 0
	WindowRect.right=(long)width;									// Set Right Value To Requested Width
	WindowRect.top=(long)0;											// Set Top Value To 0
	WindowRect.bottom=(long)height;									// Set Bottom Value To Requested Height

	fullscreen=fullscreenflag;										// Set The Global Fullscreen Flag

	hInstance			= GetModuleHandle(NULL);					// Grab An Instance For Our Window
	wc.style			= CS_HREDRAW | CS_VREDRAW | CS_OWNDC;		// Redraw On Size, And Own DC For Window.
	wc.lpfnWndProc		= (WNDPROC) WndProc;						// WndProc Handles Messages
	wc.cbClsExtra		= 0;										// No Extra Window Data
	wc.cbWndExtra		= 0;										// No Extra Window Data
	wc.hInstance		= hInstance;								// Set The Instance
	wc.hIcon			= LoadIcon(NULL, IDI_WINLOGO);				// Load The Default Icon
	wc.hCursor			= LoadCursor(NULL, IDC_ARROW);				// Load The Arrow Pointer
	wc.hbrBackground	= NULL;										// No Background Required For GL
	wc.lpszMenuName		= NULL;										// We Don't Want A Menu
	wc.lpszClassName	= "OpenGL";									// Set The Class Name

	if (!RegisterClass(&wc))										// Attempt To Register The Window Class
	{
		MessageBox(NULL,"Failed To Register The Window Class.","ERROR",MB_OK|MB_ICONEXCLAMATION);
		return FALSE;												// Return FALSE
	}
	
	if (fullscreen)													// Attempt Fullscreen Mode?
	{
		DEVMODE dmScreenSettings;									// Device Mode
		memset(&dmScreenSettings,0,sizeof(dmScreenSettings));		// Makes Sure Memory's Cleared
		dmScreenSettings.dmSize=sizeof(dmScreenSettings);			// Size Of The Devmode Structure
		dmScreenSettings.dmPelsWidth	= width;					// Selected Screen Width
		dmScreenSettings.dmPelsHeight	= height;					// Selected Screen Height
		dmScreenSettings.dmBitsPerPel	= bits;						// Selected Bits Per Pixel
		dmScreenSettings.dmFields=DM_BITSPERPEL|DM_PELSWIDTH|DM_PELSHEIGHT;

		// Try To Set Selected Mode And Get Results.  NOTE: CDS_FULLSCREEN Gets Rid Of Start Bar.
		if (ChangeDisplaySettings(&dmScreenSettings,CDS_FULLSCREEN)!=DISP_CHANGE_SUCCESSFUL)
		{
			// If The Mode Fails, Offer Two Options.  Quit Or Use Windowed Mode.
			if (MessageBox(NULL,"The Requested Fullscreen Mode Is Not Supported By\nYour Video Card. Use Windowed Mode Instead?","NeHe GL",MB_YESNO|MB_ICONEXCLAMATION)==IDYES)
			{
				fullscreen=FALSE;									// Windowed Mode Selected.  Fullscreen = FALSE
			}
			else
			{
				// Pop Up A Message Box Letting User Know The Program Is Closing.
				MessageBox(NULL,"Program Will Now Close.","ERROR",MB_OK|MB_ICONSTOP);
				return FALSE;										// Return FALSE
			}
		}
	}

	if (fullscreen)													// Are We Still In Fullscreen Mode?
	{
		dwExStyle=WS_EX_APPWINDOW;									// Window Extended Style
		dwStyle=WS_POPUP;											// Windows Style
		ShowCursor(FALSE);											// Hide Mouse Pointer
	}
	else
	{
		dwExStyle=WS_EX_APPWINDOW | WS_EX_WINDOWEDGE;				// Window Extended Style
		dwStyle=WS_OVERLAPPEDWINDOW;								// Windows Style
	}

	AdjustWindowRectEx(&WindowRect, dwStyle, FALSE, dwExStyle);		// Adjust Window To True Requested Size

	// Create The Window
	if (!(hWnd=CreateWindowEx(	dwExStyle,							// Extended Style For The Window
								"OpenGL",							// Class Name
								title,								// Window Title
								dwStyle |							// Defined Window Style
								WS_CLIPSIBLINGS |					// Required Window Style
								WS_CLIPCHILDREN,					// Required Window Style
								0, 0,								// Window Position
								WindowRect.right-WindowRect.left,	// Calculate Window Width
								WindowRect.bottom-WindowRect.top,	// Calculate Window Height
								NULL,								// No Parent Window
								NULL,								// No Menu
								hInstance,							// Instance
								NULL)))								// Dont Pass Anything To WM_CREATE
	{
		KillGLWindow();												// Reset The Display
		MessageBox(NULL,"Window Creation Error.","ERROR",MB_OK|MB_ICONEXCLAMATION);
		return FALSE;												// Return FALSE
	}

	static	PIXELFORMATDESCRIPTOR pfd=								// pfd Tells Windows How We Want Things To Be
	{
		sizeof(PIXELFORMATDESCRIPTOR),								// Size Of This Pixel Format Descriptor
		1,															// Version Number
		PFD_DRAW_TO_WINDOW |										// Format Must Support Window
		PFD_SUPPORT_OPENGL |										// Format Must Support OpenGL
		PFD_DOUBLEBUFFER,											// Must Support Double Buffering
		PFD_TYPE_RGBA,												// Request An RGBA Format
		bits,														// Select Our Color Depth
		0, 0, 0, 0, 0, 0,											// Color Bits Ignored
		0,															// No Alpha Buffer
		0,															// Shift Bit Ignored
		0,															// No Accumulation Buffer
		0, 0, 0, 0,													// Accumulation Bits Ignored
		16,															// 16Bit Z-Buffer (Depth Buffer)  
		0,															// No Stencil Buffer
		0,															// No Auxiliary Buffer
		PFD_MAIN_PLANE,												// Main Drawing Layer
		0,															// Reserved
		0, 0, 0														// Layer Masks Ignored
	};
	
	if (!(hDC=GetDC(hWnd)))											// Did We Get A Device Context?
	{
		KillGLWindow();												// Reset The Display
		MessageBox(NULL,"Can't Create A GL Device Context.","ERROR",MB_OK|MB_ICONEXCLAMATION);
		return FALSE;												// Return FALSE
	}

	if (!(PixelFormat=ChoosePixelFormat(hDC,&pfd)))					// Did Windows Find A Matching Pixel Format?
	{
		KillGLWindow();												// Reset The Display
		MessageBox(NULL,"Can't Find A Suitable PixelFormat.","ERROR",MB_OK|MB_ICONEXCLAMATION);
		return FALSE;												// Return FALSE
	}

	if(!SetPixelFormat(hDC,PixelFormat,&pfd))						// Are We Able To Set The Pixel Format?
	{
		KillGLWindow();												// Reset The Display
		MessageBox(NULL,"Can't Set The PixelFormat.","ERROR",MB_OK|MB_ICONEXCLAMATION);
		return FALSE;												// Return FALSE
	}

	if (!(hRC=wglCreateContext(hDC)))								// Are We Able To Get A Rendering Context?
	{
		KillGLWindow();												// Reset The Display
		MessageBox(NULL,"Can't Create A GL Rendering Context.","ERROR",MB_OK|MB_ICONEXCLAMATION);
		return FALSE;												// Return FALSE
	}

	if(!wglMakeCurrent(hDC,hRC))									// Try To Activate The Rendering Context
	{
		KillGLWindow();												// Reset The Display
		MessageBox(NULL,"Can't Activate The GL Rendering Context.","ERROR",MB_OK|MB_ICONEXCLAMATION);
		return FALSE;												// Return FALSE
	}

	ShowWindow(hWnd,SW_SHOW);										// Show The Window
	SetForegroundWindow(hWnd);										// Slightly Higher Priority
	SetFocus(hWnd);													// Sets Keyboard Focus To The Window
	ReSizeGLScene(width, height);									// Set Up Our Perspective GL Screen

	if (!InitGL())													// Initialize Our Newly Created GL Window
	{
		KillGLWindow();												// Reset The Display
		MessageBox(NULL,"Initialization Failed.","ERROR",MB_OK|MB_ICONEXCLAMATION);
		return FALSE;												// Return FALSE
	}

	return TRUE;													// Success
}

LRESULT CALLBACK WndProc(	HWND	hWnd,							// Handle For This Window
							UINT	uMsg,							// Message For This Window
							WPARAM	wParam,							// Additional Message Information
							LPARAM	lParam)							// Additional Message Information
{
	switch (uMsg)													// Check For Windows Messages
	{
		case WM_CREATE:
		{
			//hWndStatus = CreateStatusWindow(WS_CHILD | WS_BORDER | WS_VISIBLE, "", hWnd, (UINT)100);
		}
		case WM_ACTIVATE:											// Watch For Window Activate Message
		{
			if (!HIWORD(wParam))									// Check Minimization State
			{
				active=TRUE;										// Program Is Active
			}
			else
			{
				active=FALSE;										// Program Is No Longer Active
			}

			return 0;												// Return To The Message Loop
		}

		case WM_SYSCOMMAND:											// Intercept System Commands
		{
			switch (wParam)											// Check System Calls
			{
				case SC_SCREENSAVE:									// Screensaver Trying To Start?
				case SC_MONITORPOWER:								// Monitor Trying To Enter Powersave?
				return 0;											// Prevent From Happening
			}
			break;													// Exit
		}

		case WM_CLOSE:												// Did We Receive A Close Message?
		{
			PostQuitMessage(0);										// Send A Quit Message
			return 0;												// Jump Back
		}

		case WM_KEYDOWN:											// Is A Key Being Held Down?
		{
			keys[wParam] = TRUE;									// If So, Mark It As TRUE
			return 0;												// Jump Back
		}

		case WM_KEYUP:												// Has A Key Been Released?
		{
			keys[wParam] = FALSE;									// If So, Mark It As FALSE
			return 0;												// Jump Back
		}

		case WM_SIZE:												// Resize The OpenGL Window
		{
			//SendMessage(hWndStatus, WM_SIZE, 0, 0);
			ReSizeGLScene(LOWORD(lParam),HIWORD(lParam));			// LoWord=Width, HiWord=Height
			return 0;												// Jump Back
		}
	}

	// Pass All Unhandled Messages To DefWindowProc
	return DefWindowProc(hWnd,uMsg,wParam,lParam);
}

int WINAPI WinMain(	HINSTANCE	hInstance,							// Instance
					HINSTANCE	hPrevInstance,						// Previous Instance
					LPSTR		lpCmdLine,							// Command Line Parameters
					int			nCmdShow)							// Window Show State
{
	MSG		msg;													// Windows Message Structure
	BOOL	done=FALSE;												// Bool Variable To Exit Loop

	MAZEGAME->LoadMaze("Data/Mazes/Maze01.mz");						// Load the maze file
	MAZEGAME->BuildMenus();											// Construct menus
	MAZEGAME->triggerMenu();										// Go to menu mode to start with
	MAZEGAME->loaded = true;										// Signal that maze has been loaded

	tJoystick = new Joystick();

	pos.x = MAZEGAME->tPlayer->x;									// Initialize player position
	pos.y = 5.0f;
	pos.z = MAZEGAME->tPlayer->z;
	MAZEGAME->tMaze->fsystem->set3DListenerAttributes(0, &pos, 0, 0, 0);

    if(FAILED(::CoInitialize(NULL)))			// Initialize COM object for voice support
        return FALSE;
    hr = CoCreateInstance(CLSID_SpVoice, NULL, CLSCTX_ALL, IID_ISpVoice, (void **)&pVoice);
    if(SUCCEEDED(hr))
    {
        hr = pVoice->Speak(L"Welcome to the Maze Game!", 0, NULL);
		hr = pVoice->WaitUntilDone(1000);
		hr = pVoice->Speak(L"Menu Mode", 0, NULL);
		hr = pVoice->WaitUntilDone(1000);
		hr = pVoice->Speak(MAZEGAME->tMenus[MAZEGAME->currentMenu].menuItems[MAZEGAME->tMenus[MAZEGAME->currentMenu].selectedItem].wtext, 0, NULL);
	}

	FMOD_RESULT result;

	//if (MessageBox(NULL,"Would You Like To Run In Fullscreen Mode?", "Start FullScreen?",MB_YESNO|MB_ICONQUESTION)==IDNO)
	//{
		fullscreen=FALSE;											// Windowed Mode
	//}

	// Create Our OpenGL Window
	if (!CreateGLWindow("The Maze Project v1.0",640,480,16,fullscreen))
	{
		return 0;													// Quit If Window Was Not Created
	}

	if(tJoystick->Init(hWnd))										// Is there a joystick present?
		use_joystick = true;

	while(!done)													// Loop That Runs While done=FALSE
	{
		if (PeekMessage(&msg,NULL,0,0,PM_REMOVE))					// Is There A Message Waiting?
		{
			if (msg.message==WM_QUIT)								// Have We Received A Quit Message?
			{
				done=TRUE;											// If So done=TRUE
			}
			else													// If Not, Deal With Window Messages
			{
				TranslateMessage(&msg);								// Translate The Message
				DispatchMessage(&msg);								// Dispatch The Message
			}
		}
		else														// If There Are No Messages
		{
			// Draw The Scene.  Watch For Quit Messages From DrawGLScene()
			if ((active && !DrawGLScene()))							// Active?  Was There A Quit Received?
			{
				done=TRUE;											// ESC or DrawGLScene Signalled A Quit
			}
			else													// Not Time To Quit, Update Screen
			{
				SwapBuffers(hDC);									// Swap Buffers (Double Buffering)
			}

			/*if (keys[VK_F1])										// Is F1 Being Pressed?
			{
				keys[VK_F1]=FALSE;									// If So Make Key FALSE
				KillGLWindow();										// Kill Our Current Window
				fullscreen=!fullscreen;								// Toggle Fullscreen / Windowed Mode
				// Recreate Our OpenGL Window
				if (!CreateGLWindow("The Maze Project v1.0",640,480,16,fullscreen))
				{
					return 0;										// Quit If Window Was Not Created
				}
			}*/
			if(keys[VK_SPACE])										// KEY HANDLER
			{
				keys[VK_SPACE]=FALSE;
				MAZEGAME->switchCam();
			}
			if(keys[VK_UP])
			{
				if(MAZEGAME->menumode)
				{
					keys[VK_UP]=FALSE;
					MAZEGAME->tMenus[MAZEGAME->currentMenu].incrementMenu();
					if(SUCCEEDED(hr)) {
						hr = pVoice->WaitUntilDone(1000);
						hr = pVoice->Speak(MAZEGAME->tMenus[MAZEGAME->currentMenu].menuItems[MAZEGAME->tMenus[MAZEGAME->currentMenu].selectedItem].wtext, 0, NULL);
					}
				}
				else if(!use_joystick)
				{
					if(!MAZEGAME->DetectCollision((float)-1*sin(MAZEGAME->tPlayer->heading*pi_180) * 0.85f, 0.0f))
						MAZEGAME->tPlayer->x -= (float)sin(MAZEGAME->tPlayer->heading*pi_180) * 0.85f;
					if(!MAZEGAME->DetectCollision(0.0f, (float)-1*cos(MAZEGAME->tPlayer->heading*pi_180) * 0.85f))
						MAZEGAME->tPlayer->z -= (float)cos(MAZEGAME->tPlayer->heading*pi_180) * 0.85f;
					pos.x = MAZEGAME->tPlayer->x;
					pos.y = 5.0f;
					pos.z = MAZEGAME->tPlayer->z;
					MAZEGAME->tMaze->fsystem->set3DListenerAttributes(0, &pos, 0, 0, 0);

					if (MAZEGAME->tPlayer->walkbiasangle >= 359.0f)
					{
						MAZEGAME->tPlayer->walkbiasangle = 0.0f;
					}
					else
					{
						MAZEGAME->tPlayer->walkbiasangle+= 5;
					}
					MAZEGAME->tPlayer->walkbias = (float)sin(MAZEGAME->tPlayer->walkbiasangle * pi_180)/2.0f;
				}
			}
			if(keys[VK_DOWN])
			{
				if(MAZEGAME->menumode)
				{
					keys[VK_DOWN]=FALSE;
					MAZEGAME->tMenus[MAZEGAME->currentMenu].decrementMenu();
					if(SUCCEEDED(hr)) {
						hr = pVoice->WaitUntilDone(1000);
						hr = pVoice->Speak(MAZEGAME->tMenus[MAZEGAME->currentMenu].menuItems[MAZEGAME->tMenus[MAZEGAME->currentMenu].selectedItem].wtext, 0, NULL);
					}
				}
				else if(!use_joystick)
				{
					if(!MAZEGAME->DetectCollision((float)sin(MAZEGAME->tPlayer->heading*pi_180) * 0.85f, 0.0f))
						MAZEGAME->tPlayer->x += (float)sin(MAZEGAME->tPlayer->heading*pi_180) * 0.85f;
					if(!MAZEGAME->DetectCollision(0.0f, (float)cos(MAZEGAME->tPlayer->heading*pi_180) * 0.85f))
						MAZEGAME->tPlayer->z += (float)cos(MAZEGAME->tPlayer->heading*pi_180) * 0.85f;
					pos.x = MAZEGAME->tPlayer->x;
					pos.y = 5.0f;
					pos.z = MAZEGAME->tPlayer->z;
					MAZEGAME->tMaze->fsystem->set3DListenerAttributes(0, &pos, 0, 0, 0);

					if (MAZEGAME->tPlayer->walkbiasangle <= 1.0f)
					{
						MAZEGAME->tPlayer->walkbiasangle = 359.0f;
					}
					else
					{
						MAZEGAME->tPlayer->walkbiasangle-= 5;
					}
					MAZEGAME->tPlayer->walkbias = (float)sin(MAZEGAME->tPlayer->walkbiasangle * pi_180)/2.0f;
				}
			}
			if(keys[VK_LEFT])
			{
				if(!MAZEGAME->menumode && !use_joystick)
				{
					MAZEGAME->tPlayer->heading += 2.0f;
					float x_d =(float)sin(MAZEGAME->tPlayer->heading*pi_180);
					float z_d =(float)cos(MAZEGAME->tPlayer->heading*pi_180);
					forward.x = x_d;
					forward.y = 0.0f;
					forward.z = z_d;
					Normalize(forward.x, forward.y, forward.z);
					MAZEGAME->tMaze->fsystem->set3DListenerAttributes(0, 0, 0, &forward, 0);
				}
			}
			if(keys[VK_RIGHT])
			{
				if(!MAZEGAME->menumode && !use_joystick)
				{
					MAZEGAME->tPlayer->heading -= 2.0f;
					float x_d =(float)sin(MAZEGAME->tPlayer->heading*pi_180);
					float z_d =(float)cos(MAZEGAME->tPlayer->heading*pi_180);
					forward.x = x_d;
					forward.y = 0.0f;
					forward.z = z_d;
					Normalize(forward.x, forward.y, forward.z);
					MAZEGAME->tMaze->fsystem->set3DListenerAttributes(0, 0, 0, &forward, 0);
				}
			}
			if(keys[13]) // Enter
			{
				keys[13]=FALSE;
				if(MAZEGAME->menumode && MAZEGAME->currentMenu == 0)
				{
					if(MAZEGAME->MenuHandler(MAZEGAME->tMenus[0].selectedItem) == -1)
					{
						if(SUCCEEDED(hr)) {
							hr = pVoice->Speak(L"Goodbye", 0, NULL);
						}
						done=TRUE;
					}
					else if(MAZEGAME->MenuHandler(MAZEGAME->tMenus[0].selectedItem) == 0)
					{
						MAZEGAME->switchCam();
						MAZEGAME->triggerMenu();
						if(SUCCEEDED(hr)) {
							hr = pVoice->WaitUntilDone(1000);
							hr = pVoice->Speak(L"Starting Game", 0, NULL);
							hr = pVoice->WaitUntilDone(1000);
							hr = pVoice->Speak(L"The object of the game is to collect all of the items in the maze.", 0, NULL);
							hr = pVoice->WaitUntilDone(1000);
							hr = pVoice->Speak(L"Ready?", 0, NULL);
							hr = pVoice->WaitUntilDone(1000);
							hr = pVoice->Speak(L"Three!", 0, NULL);
							hr = pVoice->WaitUntilDone(1000);
							hr = pVoice->Speak(L"Two!", 0, NULL);
							hr = pVoice->WaitUntilDone(1000);
							hr = pVoice->Speak(L"One!", 0, NULL);
							hr = pVoice->WaitUntilDone(1000);
							hr = pVoice->Speak(L"Go!", 0, NULL);
						}
						for(int i = 0; i < MAZEGAME->tMaze->fsound_objects.size(); i++)
							result = MAZEGAME->tMaze->fsound_objects[i].channel->setPaused(false);
					}
				}
				else if(MAZEGAME->menumode && MAZEGAME->currentMenu == 1)
				{
					if(MAZEGAME->MenuHandler(MAZEGAME->tMenus[1].selectedItem) == -1) {
						MAZEGAME->tMaze->difficulty = 1;
						MAZEGAME->tMenus[1].menuItems[1].text = "Difficulty: Moderate";
					}
					else if(MAZEGAME->MenuHandler(MAZEGAME->tMenus[1].selectedItem) == 0) {
						MAZEGAME->tMaze->difficulty = 2;
						MAZEGAME->tMenus[1].menuItems[1].text = "Difficulty: Hard";
					}
					else if(MAZEGAME->MenuHandler(MAZEGAME->tMenus[1].selectedItem) == 1) {
						MAZEGAME->tMaze->difficulty = 0;
						MAZEGAME->tMenus[1].menuItems[1].text = "Difficulty: Easy";
					}
				}
			}
			if(keys[VK_ESCAPE])
			{
				keys[VK_ESCAPE]=FALSE;
				if(MAZEGAME->menumode)
				{
					if(MAZEGAME->currentMenu==1)
						MAZEGAME->currentMenu = 0;
				}
				else
				{
					for(int i = 0; i < MAZEGAME->tMaze->fsound_objects.size(); i++)
						result = MAZEGAME->tMaze->fsound_objects[i].channel->setPaused(true);
					MAZEGAME->triggerMenu();
					if(SUCCEEDED(hr)) {
						hr = pVoice->WaitUntilDone(1000);
						hr = pVoice->Speak(L"Menu Mode", 0, NULL);
						hr = pVoice->WaitUntilDone(1000);
						hr = pVoice->Speak(MAZEGAME->tMenus[MAZEGAME->currentMenu].menuItems[MAZEGAME->tMenus[MAZEGAME->currentMenu].selectedItem].wtext, 0, NULL);
					}
				}
			}
			if(keys['E'])
			{
				keys['E']=FALSE;
				MAZEGAME->showstats = true;
			}
			if(keys['F'])
			{
				keys['F']=FALSE;
				if(MAZEGAME->showFMODGeom)
					MAZEGAME->showFMODGeom = false;
				else
					MAZEGAME->showFMODGeom = true;
			}
		}
	}

	// Shutdown
	::CoUninitialize();
	KillGLWindow();													// Kill The Window
	return (msg.wParam);											// Exit The Program
}