#ifndef _MAZEAPP_H
#define _MAZEAPP_H

#include <math.h>
#include <time.h>
#include "Menu.h"
#include "Maze.h"
#include "Camera.h"
#include "Player.h"
#include "Random.h"

void Normalize(float &a, float &b, float &c);

class MazeApp
{
public:

	Maze *tMaze;
	vector<Camera> tCameras;
	vector<Menu> tMenus;
	Player *tPlayer;
	
	MazeApp();

	int currentMenu;
	int currentCamera;
	int currentTile_x, currentTile_z;
	bool menumode;
	//bool showstats;
	bool showFMODGeom;
	bool loaded;
	bool completed;
	bool ended;

	int num_collisions;

	int start_x, start_z, start_subx, start_subz;
	int end_x, end_z, end_subx, end_subz;

	bool LoadMaze(char *filename);
	bool BuildMenus();
	void Draw();
	void DrawFMODGeometry();

	FMOD::Channel *extra_channel;
	FMOD::Channel *cane_channel;
	bool cane_tapped;

	void triggerMenu();
	void switchCam();
	int MenuHandler(int id);

	bool DetectCollision(float incr_x, float incr_z);
	int getPlayerPos_X();
	int getPlayerPos_Z();
	float getPlayerLocal_X();
	float getPlayerLocal_Z();
	int getPos_X(float x, float z);
	int getPos_Z(float x, float z);
	float getLocal_X(float x, float z);
	float getLocal_Z(float x, float z);

	void UpdateObjects();

private:

	bool DetectCollisionTile(int x, int z, float lx, float ly, float lz, float incr_x, float incr_z);

	vector<float> mvec;		// variables for collision detection (could be made much more efficient)
	vector<float> norm;
	vector<float> o_p;
	vector<float> a;
	vector<float> b;
	vector<float> c;
	vector<float> m;
	vector<float> n;
	vector<float> h;
	vector<float> s;
	vector<float> q;
};

MazeApp::MazeApp()
{
	menumode = false;
	loaded = false;
	showFMODGeom = false;
	currentMenu = 0;
	currentCamera = 0;
	currentTile_x = 0;
	currentTile_z = 0;
	num_collisions = 0;
	cane_tapped = false;
	completed = false; ended = false;
	start_x = 0;
	start_z = 0;
	tMaze = new Maze();
	tPlayer = new Player();
	Camera isoCam(150.0f, 175.0f, 150.0f, 0.0f, 0.0f, 0.0f);
	Camera walkCam(0.0f, 0.0f, 0.0f, 0.0f, 0.0f, -10.0f);
	tCameras.push_back(isoCam);
	tCameras.push_back(walkCam);
	
	for(int i = 0; i < 3; i++)
	{
		mvec.push_back(0.0f);
		norm.push_back(0.0f);
		o_p.push_back(0.0f);
		a.push_back(0.0f);
		b.push_back(0.0f);
		c.push_back(0.0f);
		m.push_back(0.0f);
		n.push_back(0.0f);
		h.push_back(0.0f);
		s.push_back(0.0f);
		q.push_back(0.0f);
	}
}

bool MazeApp::LoadMaze(char *filename)
{ 
	if(tMaze->LoadMaze(filename))
	{
		start_x = tMaze->st_x;
		start_z = tMaze->st_z;
		start_subx = tMaze->st_subt_x;
		start_subz = tMaze->st_subt_z;
		end_x = tMaze->ed_x;
		end_z = tMaze->ed_z;
		end_subx = tMaze->ed_subt_x;
		end_subz = tMaze->ed_subt_z;
		tPlayer->x = (start_x-1)*39.0f+(start_subx-1)*13.0f-12.0f;
		tPlayer->z = (start_z-1)*39.0f+(start_subz-1)*13.0f-12.0f;

		FMOD_RESULT result;
	
		for(int i = 0; i < tMaze->fsound_objects.size(); i++)   // create soundobjects in FMOD
		{
			FMOD_VECTOR pos = { (tMaze->fsound_objects[i].tile_x-1)*39.0f+(tMaze->fsound_objects[i].pos_x-1)*13.0f-12.0f, 5.0f, (tMaze->fsound_objects[i].tile_z-1)*39.0f+(tMaze->fsound_objects[i].pos_z-1)*13.0f-12.0f };
			FMOD_VECTOR vel = { 0.0f, 0.0f, 0.0f };
	
			result = tMaze->fsystem->playSound(FMOD_CHANNEL_FREE, tMaze->fsounds[tMaze->fsound_objects[i].soundID], true, &tMaze->fsound_objects[i].channel);
			if(result != FMOD_OK)
			    MessageBox(NULL,"Could not play sound.","Error",MB_OK);
			result = tMaze->fsound_objects[i].channel->setMode(FMOD_LOOP_NORMAL);
			result = tMaze->fsound_objects[i].channel->set3DAttributes(&pos, &vel);
			result = tMaze->fsound_objects[i].channel->setPaused(true);
		}
		return true;
	}
	else
		return false;
}

bool MazeApp::BuildMenus()
{
	Menu MainMenu;
	MainMenu.AddMenuItem("The Maze Game v1.0", L"The Maze Game v1.0",0.0f, 2.0f, false);
	MainMenu.AddMenuItem("Start Game", L"Start Game",0.0f, 0.0f, true);
	MainMenu.AddMenuItem("Options", L"Options",0.0f, -1.0f, true);
	MainMenu.AddMenuItem("Exit", L"Exit",0.0f, -2.0f, true);
	tMenus.push_back(MainMenu);

	Menu OptionsMenu;
	OptionsMenu.AddMenuItem("Options", L"Options",0.0f, 2.0f, false);
	OptionsMenu.AddMenuItem("Difficulty: Moderate", L"Difficulty: Moderate", 0.0f, 0.0f, true);
	tMenus.push_back(OptionsMenu);

	return true;
}

void MazeApp::Draw()
{
	if(loaded)
	{
	if(!menumode)
	{
		gluLookAt(tCameras[currentCamera].pos_x, tCameras[currentCamera].pos_y, tCameras[currentCamera].pos_z,
				  tCameras[currentCamera].tar_x, tCameras[currentCamera].tar_y, tCameras[currentCamera].tar_z, 0, 1, 0);

		GLfloat xtrans = -tPlayer->x;
		GLfloat ytrans = -tPlayer->walkbias;
		GLfloat ztrans = -tPlayer->z;
		GLfloat roty = 360.0f - tPlayer->heading;

		glPushMatrix();
		if(currentCamera != 0)
		{
			glRotatef(roty, 0.0f, 1.0f, 0.0f);
			glTranslatef(xtrans, ytrans, ztrans);
			glTranslatef(0, -7.0f, 0);
			glEnable(GL_FOG);
		}
		else
		{
			glDisable(GL_FOG);
			glTranslatef(-(39*(tMaze->dim_x))/2, 0.0f, -(39*(tMaze->dim_y))/2);
		}
		tMaze->DrawMaze(getPlayerPos_X(), getPlayerPos_Z());
		glPopMatrix();

		glPushMatrix();
		if(showFMODGeom)
			DrawFMODGeometry();
		glPopMatrix();
		
		if(currentCamera == 0)
		{
			glPushMatrix();
			glTranslatef(-xtrans, 7.0f, -ztrans);
			glTranslatef(-(39*(tMaze->dim_x))/2, 0.0f, -(39*(tMaze->dim_y))/2);
			glRotatef(-roty, 0.0f, 1.0f, 0.0f);
			glScalef(4.0f, 4.0f, 4.0f);
			glEnable(GL_COLOR_MATERIAL);
			glColor3f(1.0f, 0.0f, 0.0f);
			tPlayer->Draw();
			glDisable(GL_COLOR_MATERIAL);
			glPopMatrix();
		}
	}
	else
	{
		gluLookAt(0,0,10,0,0,0,0,1,0);
		tMenus[currentMenu].DrawMenu();
	}
	}
}

void MazeApp::DrawFMODGeometry()
{
	int num_polys = 0;
	FMOD_RESULT result;
	FMOD_VECTOR temp_a, temp_b, temp_c;
	result = tMaze->fgeometry->getNumPolygons(&num_polys);
	glTranslatef(-(39*(tMaze->dim_x))/2, 0.0f, -(39*(tMaze->dim_y))/2);
	for(int x = 0; x < num_polys; x++)
	{
		glBegin(GL_LINES);
			result = tMaze->fgeometry->getPolygonVertex(x, 0, &temp_a);
			result = tMaze->fgeometry->getPolygonVertex(x, 1, &temp_b);
			result = tMaze->fgeometry->getPolygonVertex(x, 2, &temp_c);
			glVertex3f(temp_a.x, temp_a.y, temp_a.z);
			glVertex3f(temp_b.x, temp_b.y, temp_b.z);
			glVertex3f(temp_b.x, temp_b.y, temp_b.z);
			glVertex3f(temp_c.x, temp_c.y, temp_c.z);
			glVertex3f(temp_c.x, temp_c.y, temp_c.z);
			glVertex3f(temp_a.x, temp_a.y, temp_a.z);
		glEnd();
	}
}

void MazeApp::triggerMenu()
{ 
	if(menumode)
	{
		glClearColor(0.5f,0.5f,1.0f,1.0f);
		glDisable(GL_COLOR_MATERIAL);
		menumode = false;
	}
	else
	{
		glClearColor(0.0f,0.0f,0.0f,1.0f);
		glEnable(GL_COLOR_MATERIAL);
		tMenus[currentMenu].selectedItem = 1;
		menumode = true;
	}
}

void MazeApp::switchCam()
{
	currentCamera = (currentCamera+1)%tCameras.size();
}

int MazeApp::MenuHandler(int id)
{
	if(currentMenu == 0)
	{
		if(id == 3)
			return -1;
		else if(id == 2)
			currentMenu = 1;
		else
			return 0;
	}
	else if(currentMenu == 1)
	{
		if(id == 1 && tMenus[1].menuItems[1].text == "Difficulty: Hard")
			return -1;
		else if(id == 1 && tMenus[1].menuItems[1].text == "Difficulty: Easy")
			return 0;
		else if(id == 1 && tMenus[1].menuItems[1].text == "Difficulty: Moderate")
			return 1;
	}

	return 0;
}

int MazeApp::getPlayerPos_X()
{
	return ceil((tPlayer->x)/39.0f-0.5f);
}

int MazeApp::getPlayerPos_Z()
{
	return ceil((tPlayer->z)/39.0f-0.5f);
}

float MazeApp::getPlayerLocal_X()
{
	return (float)tPlayer->x-39.0f*getPlayerPos_X();
}

float MazeApp::getPlayerLocal_Z()
{
	return (float)tPlayer->z-39.0f*getPlayerPos_Z();
}

int MazeApp::getPos_X(float x, float z)
{
	return ceil(x/39.0f-0.5f);
}

int MazeApp::getPos_Z(float x, float z)
{
	return ceil(z/39.0f-0.5f);
}

float MazeApp::getLocal_X(float x, float z)
{
	return (float)x-39.0f*getPos_X(x,z);
}

float MazeApp::getLocal_Z(float x, float z)
{
	return (float)z-39.0f*getPos_Z(x,z);
}

bool MazeApp::DetectCollision(float incr_x, float incr_z)
{
	int x = getPlayerPos_X();
	int z = getPlayerPos_Z();
	float lx = getPlayerLocal_X();
	float ly = 5.0f;
	float lz = getPlayerLocal_Z();

	bool a = DetectCollisionTile(x+1,z,lx-39.0f,ly,lz,incr_x,incr_z);
	bool b = DetectCollisionTile(x,z+1,lx,ly,lz-39.0f,incr_x,incr_z);
	bool c = DetectCollisionTile(x,z,lx,ly,lz,incr_x,incr_z);
	bool d = DetectCollisionTile(x-1,z,lx+39.0f,ly,lz,incr_x,incr_z);
	bool e = DetectCollisionTile(x,z-1,lx,ly,lz+39.0f,incr_x,incr_z);

	if(a || b || c || d || e)
		return true;

	return false;
}

bool MazeApp::DetectCollisionTile(int x, int z, float lx, float ly, float lz, float incr_x, float incr_z)
{
	bool no_intersect = false;
	float g, f, u, v;
	float intersect = 0.0f;
	float numer, denom;
	int rot = 0;

	if(x<=tMaze->dim_x && x>=-1 && z<=tMaze->dim_y && z>=-1)
	{	
		MilkshapeModel *cTile = new MilkshapeModel();

		if(x==tMaze->dim_x) { x=tMaze->dim_x-1; lx += 39.0f; }
		if(x==-1) { x=0; lx -= 39.0f; }
		if(z==tMaze->dim_y) { z=tMaze->dim_y-1; lz += 39.0f; }
		if(z==-1) { z=0; lz -= 39.0f; }

		cTile = tMaze->tTiles[tMaze->tGrid->getTileID(x,z)-1].tModel;
		rot = tMaze->tGrid->getTileRot(x,z);

		float temp = 0.0f;
		if(rot==2) { temp = lx; lx = lz; lz = temp; lz = -lz; temp = incr_x; incr_x = incr_z; incr_z = temp; incr_z = -incr_z; }
		if(rot==3) { lx = -lx; incr_x = -incr_x; lz = -lz; incr_z = -incr_z; }
		if(rot==4) { temp = lx; lx = lz; lz = temp; lx = -lx; temp = incr_x; incr_x = incr_z; incr_z = temp; incr_x = -incr_x; }

		mvec[0] = incr_x*6.0f;
		mvec[1] = 0.0f;
		mvec[2] = incr_z*6.0f;
	
		for(int r = 0; r < cTile->m_numTriangles; r++)
		{
			a[0] = cTile->m_pVertices[cTile->m_pTriangles[r].m_vertexIndices[0]].m_location[0];
			a[1] = cTile->m_pVertices[cTile->m_pTriangles[r].m_vertexIndices[0]].m_location[1];
			a[2] = cTile->m_pVertices[cTile->m_pTriangles[r].m_vertexIndices[0]].m_location[2];
			b[0] = cTile->m_pVertices[cTile->m_pTriangles[r].m_vertexIndices[1]].m_location[0];
			b[1] = cTile->m_pVertices[cTile->m_pTriangles[r].m_vertexIndices[1]].m_location[1];
			b[2] = cTile->m_pVertices[cTile->m_pTriangles[r].m_vertexIndices[1]].m_location[2];
			c[0] = cTile->m_pVertices[cTile->m_pTriangles[r].m_vertexIndices[2]].m_location[0];
			c[1] = cTile->m_pVertices[cTile->m_pTriangles[r].m_vertexIndices[2]].m_location[1];
			c[2] = cTile->m_pVertices[cTile->m_pTriangles[r].m_vertexIndices[2]].m_location[2];
			n[0] = b[0]-a[0];
			n[1] = b[1]-a[1];
			n[2] = b[2]-a[2];
			m[0] = c[0]-a[0];
			m[1] = c[1]-a[1];
			m[2] = c[2]-a[2];
			norm[0] = m[1]*n[2]-m[2]*n[1];
			norm[1] = m[2]*n[0]-m[0]*n[2];
			norm[2] = m[0]*n[1]-m[1]*n[0];
			Normalize(norm[0], norm[1], norm[2]);
		
			o_p[0] = lx-a[0];
			o_p[1] = ly-a[1];
			o_p[2] = lz-a[2];
			numer = (o_p[0]*norm[0])+(o_p[1]*norm[1])+(o_p[2]*norm[2]);
			denom = (mvec[0]*norm[0])+(mvec[1]*norm[1])+(mvec[2]*norm[2]);
			if(denom!=0.0f)
				intersect = -1*(numer/denom);

			if(intersect > 0.0f && intersect <= 1.0f) // intersected the plane
			{
				no_intersect = false;
				num_collisions++;

				h[0] = mvec[1]*m[2]-mvec[2]*m[1];
				h[1] = mvec[2]*m[0]-mvec[0]*m[2];
				h[2] = mvec[0]*m[1]-mvec[1]*m[0];
				g = (n[0]*h[0])+(n[1]*h[1])+(n[2]*h[2]);

				if(g > -0.00001 && g < 0.00001)
					no_intersect = true;
				else
				{
					f = 1 / g;
					s[0] = lx-a[0];
					s[1] = ly-a[1];
					s[2] = lz-a[2];
					u = f*((s[0]*h[0])+(s[1]*h[1])+(s[2]*h[2]));

					if(u < 0.0 || u > 1.0)
						no_intersect = true;
					else
					{
						q[0] = s[1]*n[2]-s[2]*n[1];
						q[1] = s[2]*n[0]-s[0]*n[2];
						q[2] = s[0]*n[1]-s[1]*n[0];
						v = f*((mvec[0]*q[0])+(mvec[1]*q[1])+(mvec[2]*q[2]));

						if(v < 0.0 || u + v > 1.0)
							no_intersect = true;
					}
				}

				if(!no_intersect)
					return true;
			}
		}
	}
	return false;
}


void MazeApp::UpdateObjects()
{
	FMOD_RESULT result;
	completed = true;
	for(int i = 0; i < tMaze->num_objects; i++)
	{
		float obj_x = tMaze->fsound_objects[i].abs_x;
		float obj_z = tMaze->fsound_objects[i].abs_z;
		if(tPlayer->x > obj_x-3.0f && tPlayer->x < obj_x+3.0f && tPlayer->z > obj_z-3.0f && tPlayer->z < obj_z+3.0f)
		{
			if(!tMaze->fsound_objects[i].hidden)
				result = tMaze->fsystem->playSound(FMOD_CHANNEL_FREE, tMaze->fsounds[4], false, &extra_channel);
			tMaze->fsound_objects[i].hidden = true;
			tMaze->fsound_objects[i].channel->setPaused(true);
		}
		if(!tMaze->fsound_objects[i].hidden)
			completed = false;
	}

	if(completed && !ended)
	{
		result = tMaze->fsystem->playSound(FMOD_CHANNEL_FREE, tMaze->fsounds[3], false, &extra_channel);
		ended = true;
	}
	
	FMOD_VECTOR pos = { tPlayer->x+(sin(tPlayer->heading)*tPlayer->walkbias*4.0f), 5.0f, tPlayer->z+(cos(tPlayer->heading)*tPlayer->walkbias*4.0f) };
	result = cane_channel->set3DAttributes(&pos, 0);
	if(tPlayer->walkbias < -0.4f || tPlayer->walkbias > 0.4f)
	{
		if(!cane_tapped)
		{
			cane_tapped = true;
			result = tMaze->fsystem->playSound(FMOD_CHANNEL_FREE, tMaze->fsounds[1], false, &cane_channel);
			if(result != FMOD_OK)
				MessageBox(NULL,"Could not play sound.","Error",MB_OK);
		}
	}
	else
	{
		if(cane_tapped)
			cane_tapped = false;
	}
}

void Normalize(float &a, float &b, float &c)
{
	float m = sqrt(a*a+b*b+c*c);
	a=a/m; b=b/m; c=c/m;
}


#endif