#ifndef _MENU_H
#define _MENU_H

#include <windows.h>					// Header File For Windows
#include <math.h>						// Header File For Windows Math Library
#include <stdio.h>						// Header File For Standard Input/Output
#include <stdarg.h>						// Header File For Variable Argument Routines
#include <vector>
using namespace std;

#include "Main.h"

GLvoid glPrint(const char *fmt, ...);

class Menu
{
public:

	GLuint base;

	struct MenuItem
	{
		float *position;
		char *text;
		WCHAR *wtext;				// for voice driver (could be changed?)
		bool selectable;
	};

	int numItems;
	int selectedItem;
	vector<MenuItem> menuItems;
	
	Menu();

	void DrawMenu();
	void AddMenuItem(char *item, WCHAR *witem, float x, float y, bool select);
	void incrementMenu();
	void decrementMenu();
};

Menu::Menu()
{
	numItems = 0;
	selectedItem = 1;
}

void Menu::DrawMenu()
{
	for(int i = 0; i < numItems; i++)
	{
		glPushMatrix();
		if(i == selectedItem && menuItems[i].selectable) 
			glColor3f(0.8f, 0.0f, 0.0f);
		else if(i != selectedItem && menuItems[i].selectable)
			glColor3f(0.8f, 0.8f, 0.8f);
		else if(!menuItems[i].selectable)
			glColor3f(1.0f, 1.0f, 1.0f);
		glTranslatef(menuItems[i].position[0], menuItems[i].position[1], 0.0f);
		glPrint(menuItems[i].text);
		glPopMatrix();
	}
}

void Menu::AddMenuItem(char *item, WCHAR *witem, float x, float y, bool select)
{
	MenuItem temp;
	temp.text = item;
	temp.wtext = witem;
	temp.position = new float[2];
	temp.position[0] = x;
	temp.position[1] = y;
	temp.selectable = select;
	menuItems.push_back(temp);
	numItems++;
}

void Menu::decrementMenu()
{
	selectedItem = (selectedItem+1)%numItems;
	if(!menuItems[selectedItem].selectable)
		selectedItem = (selectedItem+1)%numItems;
}

void Menu::incrementMenu()
{
	if(selectedItem == 0)
		selectedItem = numItems-1;
	else
		selectedItem--;
	if(!menuItems[selectedItem].selectable)
	{
		if(selectedItem == 0)
			selectedItem = numItems-1;
		else
			selectedItem--;
	}
}

#endif